#include "ac.h"
#include "ac_dirent.h"
#include "ac_stdio.h"
#include "ac_url.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/dirent.h>
#include <libxml/parser.h>
#include <libxml/tree.h>

AC_DIR *ac_opendir(const char *url) {
	char *localurl;
	AC_DIR *d = (AC_DIR *)malloc(sizeof(AC_DIR));
	if (!d) return NULL;
	d->curdirent = NULL;
	d->curchar = 0;
	if (ac_url_islocal(url)) {
		d->flags = AC_DIRENT_FLAG_LOCAL;
		localurl = ac_url_makelocal(url);
		AC_DEBUG("opening %s in local mode", localurl);
		d->localdir = opendir(localurl);
		if (!d->localdir) {
			free(d);
			return NULL;
		}
	} else {
		d->f = ac_fopen(url, "r");
		d->list = NULL;
	}
	d->content = NULL;
	d->url = (char *)malloc(sizeof(char) * strlen(url) + 1);
	strcpy(d->url, url);
	return d;
}


static void ac_createhtmlentries(AC_DIR *d, xmlNode *cur) {
	struct htmlentry *entry;
	struct htmlentrylist *list;
	char *href;
	while (cur) {
		if (cur->name == NULL) {
			cur = cur->next;
			continue;
		}
		if ((cur->name[0] == 'a' || cur->name[0] == 'A') && cur->name[1] == '\0') {
			if (d->list == NULL) {
				/* First entry -> create and set cur to it */
				d->list = malloc(sizeof(struct htmlentrylist));
				list = d->list;
				list->head = malloc(sizeof(struct htmlentry));
				list->cur = list->head;
				entry = list->cur;
				entry->next = NULL;
			} else {
				list = d->list;
				list->cur->next = (struct htmlentry *)malloc(sizeof(struct htmlentry));
				entry = list->cur->next;
				list->cur = entry;
				entry->next = NULL;
			}
			/* Setup the entry */
			href = xmlGetProp(cur, "href");
			if (href) {
				strcpy(entry->url, href);
				strcpy(entry->name, href);
				xmlFree(href);
			} else {
				entry->url[0] = '\0';
				entry->name[0] = '\0';
			}
		} else if (cur->children) {
			/* Check the children (if they exist)*/
			ac_createhtmlentries(d, cur->children);
		}
		cur = cur->next;
	}
}

static struct ac_dirent *ac_create_dirent_from_htmlentry(struct htmlentry *entry, struct ac_dirent *de) {
	de->d_ino = 0;
	strncpy(de->d_name, entry->url, NAME_MAX);
	return de;
}

static void ac_html_parse_list(AC_DIR *d) {
	xmlDocPtr doc;
	char *buf;
	int bufsize;
	struct htmlentrylist *list;
	if (d->list == NULL) {
		/* Parse the response and build the internal list structure */
		buf = d->content;
		bufsize = d->f->offset;
		doc = xmlReadMemory(buf, bufsize, d->url, NULL, XML_PARSE_RECOVER | XML_PARSE_NOERROR);
		d->list = NULL;
		if (doc) {
			ac_createhtmlentries(d, xmlDocGetRootElement(doc));
			xmlFreeDoc(doc);
			list = d->list;
			/* We just created the list so reset it */
			list->cur = list->head;
		} else {
			/* Cannot parse the document -> create an empty list */
			d->list = NULL;
		}
	}
}

static void ac_get_and_create(AC_DIR *d) {
	int read;
	read = ac_fread_complete((void **)&(d->content), d->f);
	AC_DEBUG("read %d bytes", read);
	if (d->content != NULL && read > 0) {
		d->content[read] = '\0';
		ac_html_parse_list(d);
		free(d->content);
	} else {
		d->curdirent = NULL;
	}
}

/**
 * d->list should point to a valid htmlentrylist
 */
static void ac_internal_readdir(AC_DIR *d) {
	struct htmlentrylist *list;
	struct htmlentry *entry;
	if (d->list == NULL) {
		free(d->curdirent);
		d->curdirent = NULL;
		return;
	}
	list = d->list;
	entry = list->cur;
	if (entry != NULL) {
		list->cur = list->cur->next;
		if (!d->curdirent) {
			d->curdirent = malloc(sizeof(struct ac_dirent));
		}
		ac_create_dirent_from_htmlentry(entry, d->curdirent);
	} else {
		free(d->curdirent);
		d->curdirent = NULL;
	}
}

/**
 * This should be replaced by a more generic, extensible approach 
 */
struct ac_dirent *ac_readdir(AC_DIR *d) {
	if (d->flags & AC_DIRENT_FLAG_LOCAL) {
		return readdir(d->localdir);
	}
	if (d->list == NULL) {
		ac_get_and_create(d);
	}
	ac_internal_readdir(d);
	return d->curdirent;
}

int ac_closedir(AC_DIR *d) {
	struct htmlentrylist *list;
	struct htmlentry *entry, *nextentry;
	if (d->flags & AC_DIRENT_FLAG_LOCAL) {
		return closedir(d->localdir);
	}
	ac_fclose(d->f);
	free(d->curdirent);
	if (d->list) {
		list = d->list;
		entry = list->head;
		nextentry = entry;
		while (nextentry) {
			nextentry = entry->next;
			free(entry);
			entry = nextentry;
		}
		free(list);
	}
	free(d);
	return 0;
}
